==> populus-0.6.0.tar.gz: populus-0.6.0/populus/contracts/functions.py <==
from ethereum import utils as ethereum_utils
from ethereum import abi

from populus.contracts.common import ContractBound
from populus.contracts.utils import (
    clean_args,
)


class Function(ContractBound):
    def __init__(self, name, inputs=None, outputs=None, constant=False):
        self.name = name
        self.inputs = inputs
        self.outputs = outputs
        self.constant = constant

    def __str__(self):
        signature = "{func_name}({arg_types})".format(
            func_name=self.name,
            arg_types=', '.join(
                "{0} {1}".format(i['type'], i['name']) for i in self.inputs
            )
        )
        return signature

    def __copy__(self):
        return self.__class__(self.name, self.inputs, self.outputs, self.constant)

    def abi_args_signature(self, args):
        """
        Given the calling `args` for the function call, abi encode them.
        """
        if len(self.input_types) != len(args):
            raise ValueError("Expected {0} arguments, only got {1}".format(len(self.input_types), len(args)))  # NOQA
        scrubbed_args = tuple(clean_args(*zip(self.input_types, args)))
        return abi.encode_abi(self.input_types, scrubbed_args)

    def get_call_data(self, args):
        """
        TODO: this needs tests.
        """
        prefix = self.encoded_abi_signature
        suffix = self.abi_args_signature(args)
        data = "{0}{1}".format(prefix, suffix)
        return ethereum_utils.encode_hex(data)

    def __get__(self, obj, type=None):
        if obj is None:
            return self
        else:
            return obj._meta.functions[self.name]

    def __call__(self, *args, **kwargs):
        if self.constant:
            return self.call(*args, **kwargs)
        return self.sendTransaction(*args, **kwargs)

    def sendTransaction(self, *args, **kwargs):
        data = self.get_call_data(args)

        if 'gas' not in kwargs:
            kwargs['gas'] = self.contract._meta.blockchain_client.get_max_gas()

        return self.contract._meta.blockchain_client.send_transaction(
            to=self.contract._meta.address,
            data=data,
            **kwargs
        )

    def call(self, *args, **kwargs):
        raw = kwargs.pop('raw', False)
        data = self.get_call_data(args)

        output = self.contract._meta.blockchain_client.call(
            to=self.contract._meta.address,
            data=data,
            **kwargs
        )
        if raw:
            return output
        return self.cast_return_data(output)


from ethereum.abi import process_type


def validate_argument(arg_meta, value):
    base, sub, arrlist = process_type(arg_meta['type'])

    if base == 'int':
        if not isinstance(value, (int, long)):
            return False
        exp = int(sub)
        lower_bound = (-1 * 2 ** (exp - 1)) + 1
        upper_bound = (2 ** (exp - 1)) - 1
        return lower_bound <= value <= upper_bound
    elif base == 'uint':
        if not isinstance(value, (int, long)):
            return False
        exp = int(sub)
        lower_bound = 0
        upper_bound = (2 ** exp) - 1
        return lower_bound <= value <= upper_bound
    elif base == 'address':
        if not isinstance(value, basestring):
            return False
        _value = value[2:] if value.startswith('0x') else value
        if set(_value).difference('1234567890abcdef'):
            return False
        return len(_value) == 40
    elif base == 'bytes':
        if not isinstance(value, basestring):
            return False
        max_length = int(sub)
        return len(value) <= max_length
    else:
        raise ValueError("Unsupported base: '{0}'".format(base))


class FunctionGroup(object):
    def __init__(self, functions):
        self.functions = functions

    def __str__(self):
        return "\n".join((
            "{0}: {1}".format(i, f.signature) for i, f in enumerate(self.functions)
        ))

    @property
    def name(self):
        return self.functions[0].name

    def _bind(self, contract):
        [f._bind(contract) for f in self.functions]

    def __call__(self, *args, **kwargs):
        function = self.get_function_for_call_signature(args)
        return function(*args, **kwargs)

    def sendTransaction(self, *args, **kwargs):
        function = self.get_function_for_call_signature(args)
        return function.sendTransaction(*args, **kwargs)

    def call(self, *args, **kwargs):
        function = self.get_function_for_call_signature(args)
        return function.call(*args, **kwargs)

    def get_function_for_call_signature(self, args):
        candidates = []
        for function in self.functions:
            if len(function.inputs) != len(args):
                continue
            is_match = all((
                validate_argument(arg_meta, arg)
                for arg_meta, arg
                in zip(function.inputs, args)
            ))
            if not is_match:
                continue
            candidates.append(function)
        if len(candidates) == 1:
            return candidates[0]
        elif len(candidates) == 0:
            raise TypeError("No functions matched the calling signature")
        else:
            raise TypeError("More than one function matched.")

    def abi_args_signature(self, args):
        function = self.get_function_for_call_signature(args)
        return function.abi_args_signature(args)

    def get_call_data(self, args):
        function = self.get_function_for_call_signature(args)
        return function.get_call_data(args)

    @property
    def input_types(self):
        raise AttributeError("You must access this function on the correct sub-function")

    @property
    def signature(self):
        raise AttributeError("You must access this function on the correct sub-function")

    @property
    def abi_signature(self):
        raise AttributeError("You must access this function on the correct sub-function")

    @property
    def encoded_abi_signature(self):
        raise AttributeError("You must access this function on the correct sub-function")

    @property
    def output_types(self):
        raise AttributeError("You must access this function on the correct sub-function")

    def cast_return_data(self, outputs):
        raise AttributeError("You must access this function on the correct sub-function")

==> populus-0.6.1.tar.gz: populus-0.6.1/populus/contracts/functions.py <==
from ethereum import utils as ethereum_utils
from ethereum import abi

from populus.contracts.common import ContractBound
from populus.contracts.utils import (
    clean_args,
)

GAS_LIMIT_FRACTION = 0.9


class Function(ContractBound):
    def __init__(self, name, inputs=None, outputs=None, constant=False):
        self.name = name
        self.inputs = inputs
        self.outputs = outputs
        self.constant = constant

    def __str__(self):
        signature = "{func_name}({arg_types})".format(
            func_name=self.name,
            arg_types=', '.join(
                "{0} {1}".format(i['type'], i['name']) for i in self.inputs
            )
        )
        return signature

    def __copy__(self):
        return self.__class__(self.name, self.inputs, self.outputs, self.constant)

    def abi_args_signature(self, args):
        """
        Given the calling `args` for the function call, abi encode them.
        """
        if len(self.input_types) != len(args):
            raise ValueError("Expected {0} arguments, only got {1}".format(len(self.input_types), len(args)))  # NOQA
        scrubbed_args = tuple(clean_args(*zip(self.input_types, args)))
        return abi.encode_abi(self.input_types, scrubbed_args)

    def get_call_data(self, args):
        """
        TODO: this needs tests.
        """
        prefix = self.encoded_abi_signature
        suffix = self.abi_args_signature(args)
        data = "{0}{1}".format(prefix, suffix)
        return ethereum_utils.encode_hex(data)

    def __get__(self, obj, type=None):
        if obj is None:
            return self
        else:
            return obj._meta.functions[self.name]

    def __call__(self, *args, **kwargs):
        if self.constant:
            return self.call(*args, **kwargs)
        return self.sendTransaction(*args, **kwargs)

    def sendTransaction(self, *args, **kwargs):
        data = self.get_call_data(args)

        if 'gas' not in kwargs:
            # The gasLimit value on the geth chain seems to continuously
            # decline after every block. We use a value of gas
            # slightly less than the get_max_gas value so that when
            # the transaction gets processed, the gas value we send
            # is still less than the gasLimit value when our transaction
            # eventually gets processed.
            kwargs['gas'] = int(
                GAS_LIMIT_FRACTION * self.contract._meta.blockchain_client.get_max_gas()
            )

        return self.contract._meta.blockchain_client.send_transaction(
            to=self.contract._meta.address,
            data=data,
            **kwargs
        )

    def call(self, *args, **kwargs):
        raw = kwargs.pop('raw', False)
        data = self.get_call_data(args)

        output = self.contract._meta.blockchain_client.call(
            to=self.contract._meta.address,
            data=data,
            **kwargs
        )
        if raw:
            return output
        return self.cast_return_data(output)


from ethereum.abi import process_type


def validate_argument(arg_meta, value):
    base, sub, arrlist = process_type(arg_meta['type'])

    if base == 'int':
        if not isinstance(value, (int, long)):
            return False
        exp = int(sub)
        lower_bound = (-1 * 2 ** (exp - 1)) + 1
        upper_bound = (2 ** (exp - 1)) - 1
        return lower_bound <= value <= upper_bound
    elif base == 'uint':
        if not isinstance(value, (int, long)):
            return False
        exp = int(sub)
        lower_bound = 0
        upper_bound = (2 ** exp) - 1
        return lower_bound <= value <= upper_bound
    elif base == 'address':
        if not isinstance(value, basestring):
            return False
        _value = value[2:] if value.startswith('0x') else value
        if set(_value).difference('1234567890abcdef'):
            return False
        return len(_value) == 40
    elif base == 'bytes':
        if not isinstance(value, basestring):
            return False
        max_length = int(sub)
        return len(value) <= max_length
    else:
        raise ValueError("Unsupported base: '{0}'".format(base))


class FunctionGroup(object):
    def __init__(self, functions):
        self.functions = functions

    def __str__(self):
        return "\n".join((
            "{0}: {1}".format(i, f.signature) for i, f in enumerate(self.functions)
        ))

    @property
    def name(self):
        return self.functions[0].name

    def _bind(self, contract):
        [f._bind(contract) for f in self.functions]

    def __call__(self, *args, **kwargs):
        function = self.get_function_for_call_signature(args)
        return function(*args, **kwargs)

    def sendTransaction(self, *args, **kwargs):
        function = self.get_function_for_call_signature(args)
        return function.sendTransaction(*args, **kwargs)

    def call(self, *args, **kwargs):
        function = self.get_function_for_call_signature(args)
        return function.call(*args, **kwargs)

    def get_function_for_call_signature(self, args):
        candidates = []
        for function in self.functions:
            if len(function.inputs) != len(args):
                continue
            is_match = all((
                validate_argument(arg_meta, arg)
                for arg_meta, arg
                in zip(function.inputs, args)
            ))
            if not is_match:
                continue
            candidates.append(function)
        if len(candidates) == 1:
            return candidates[0]
        elif len(candidates) == 0:
            raise TypeError("No functions matched the calling signature")
        else:
            raise TypeError("More than one function matched.")

    def abi_args_signature(self, args):
        function = self.get_function_for_call_signature(args)
        return function.abi_args_signature(args)

    def get_call_data(self, args):
        function = self.get_function_for_call_signature(args)
        return function.get_call_data(args)

    @property
    def input_types(self):
        raise AttributeError("You must access this function on the correct sub-function")

    @property
    def signature(self):
        raise AttributeError("You must access this function on the correct sub-function")

    @property
    def abi_signature(self):
        raise AttributeError("You must access this function on the correct sub-function")

    @property
    def encoded_abi_signature(self):
        raise AttributeError("You must access this function on the correct sub-function")

    @property
    def output_types(self):
        raise AttributeError("You must access this function on the correct sub-function")

    def cast_return_data(self, outputs):
        raise AttributeError("You must access this function on the correct sub-function")
